package JElriDEV;

import java.io.*;

/*
 * (C) ELRIDEV SOFTWARE, 1999. ALL RIGHTS RESERVED.
 * THIS SOFTWARE IS THE PROPERTY OF ELRIDEV SOFTWARE, YOU CAN USE THIS SOFTWARE
 * ON YOUR WEBPAGE (OR USE THE SOURCE FOR LEARNING JAVA), IT PROVIDED AS AN EXAMPLE,
 * WITH NO WARRANTIES WHATSOEVER, SHOULD ANY DAMAGE OCCUR FROM THE USE OF THIS
 * PRODUCT ELRIDEV WILL BE HELD IN NO WAY LIABLE FOR DAMAGES.
 * YOU CAN ALSO MODIFY THIS CODE FOR YOU OWN PURPOSES, HOWEVER ONCE THIS IS DONE
 * YOU MUST INDICATE THAT THESE CHANGES HAVE BEEN MADE, SHOULD YOU WISH IT REMAIN
 * FREELY AVAILABLE.
 */

/**
 * a simple class that can be used to load binary files that were created on a
 * different platform. It has some simple byte ordering support built in.
 * This presently does not support random access, its designed for high performance
 * buffered serial access to an input stream.
 */
public class BinaryFileReader
{
    protected boolean     bError = false; // if an error has occurred
    protected InputStream input;          // the input stream
    protected int         byteOrd;        // byte order

    /** specify intel byte ordering */
    public static final int   intelOrdering=1;
    /** specify motorola byte ordering */
    public static final int   motorOrdering=2;

    // used for read ahead buffering.
    protected byte        readAhead[];
    protected int         iReadAheadPos,iReadAheadSize,iActual;

    /**
     * create an binary file reader that uses an input stream to read information.
     * the intel byte ordering is used.
     * @param is the input stream to read from
     */
    public BinaryFileReader(InputStream is)
    {
        input=is;
        byteOrd=intelOrdering;
        iReadAheadSize=5000;
    }

    /**
     * create an binary file reader that uses an input stream to read information
     * with a particular byte ordering mode
     * @param is the input stream to read from
     * @param byteOrder the byte ordering of the file being read
     */
    public BinaryFileReader(InputStream is,int byteOrder)
    {
        input=is;
        byteOrd = byteOrder;
        iReadAheadSize=5000;
    }

    /**
     * set the size of the buffer used to read ahead, by default this
     * value is set to 5,000 which works well for most files.
     * @param iSize the size of the buffer
     */
    public void setReadAheadSize(int iSize)
    {
        if(iSize>20000) iSize=20000; // we dont want too large a buffer
        iReadAheadSize=iSize;
    }

    /**
     * sets the bits indicated by shift to shift + 8 bits with the values in
     * byte parameter b. This is safe with signed values without extending the
     * sign over the 8 bit boundary
     * @param b the byte to be inserted into the int
     * @param shift how many position to move the byte
     */
    public static int intBits(byte b, int shift)
    {
        int ret=0;
        int curpt = (int)Math.pow(2,shift);
        int bypt=1;
        for(int i=0;i<8;i++)
        {
            if((b&bypt)!=0) ret|=curpt;
            curpt=curpt*2;
            bypt=bypt*2;
        }
        return ret;
    }

    /**
     * sets the bits indicated by shift to shift + 8 bits with the values in
     * byte parameter b. This is safe with signed values without extending the
     * sign over the 8 bit boundary
     * @param b the byte to be inserted into the long
     * @param shift how many position to move the byte
     */
    public static long longBits(byte b, int shift)
    {
        long ret=0;
        long curpt = (long)Math.pow(2,shift);
        int bypt=1;
        for(int i=0;i<8;i++)
        {
            if((b&bypt)!=0) ret|=curpt;
            curpt=curpt*2;
            bypt=bypt*2;
        }
        return ret;
    }

    /**
     * reads an int (32 bit) from the file
     * @return the int value
     */
    public int readInt()
    {
        byte b1 = getByte();
        byte b2 = getByte();
        byte b3 = getByte();
        byte b4 = getByte();
        if(byteOrd==intelOrdering)
        {
            return intBits(b4,24)|intBits(b3,16)|intBits(b2,8)|intBits(b1,0);
        }
        else
        {
            return intBits(b1,24)|intBits(b2,16)|intBits(b3,8)|b4;
        }
    }

    /**
     * reads an double (64 bit) from the file
     * @return the double value
     */
    public double readDouble()
    {
        byte b1 = getByte();
        byte b2 = getByte();
        byte b3 = getByte();
        byte b4 = getByte();
        byte b5 = getByte();
        byte b6 = getByte();
        byte b7 = getByte();
        byte b8 = getByte();
        long ldbl=0;
        if(byteOrd==intelOrdering)
        {
            ldbl = longBits(b8,56)|longBits(b7,48)|longBits(b6,40)|longBits(b5,32)|longBits(b4,24)|longBits(b3,16)|longBits(b2,8)|longBits(b1,0);
        }
        else
        {
            ldbl = longBits(b1,56)|longBits(b2,48)|longBits(b3,40)|longBits(b4,32)|longBits(b5,24)|longBits(b6,16)|longBits(b7,8)|longBits(b8,0);
        }
        return Double.longBitsToDouble(ldbl);
    }

    /**
     * reads an float (32 bit) from the file
     * @return the float value
     */
    public float readFloat()
    {
        byte b1 = getByte();
        byte b2 = getByte();
        byte b3 = getByte();
        byte b4 = getByte();
        int flt;
        if(byteOrd==intelOrdering)
        {
            flt = intBits(b4,24)|intBits(b3,16)|intBits(b2,8)|intBits(b1,0);
        }
        else
        {
            flt = intBits(b1,24)|intBits(b2,16)|intBits(b3,8)|intBits(b4,0);
        }
        return Float.intBitsToFloat(flt);
    }

    /**
     * reads an byte (8 bit) from the file
     * @return the byte value
     */
    public byte readByte()
    {
        return getByte();
    }

    /**
     * gets a byte from the file, reading either from the buffer or file.
     * @return the byte value
     */
    protected byte getByte()
    {
        // if we've no buffer, then allocate one, and fill it up
        if(readAhead==null)
        {
            readAhead=new byte[iReadAheadSize];
            getMoreBytes();
        }

        // check if were at the end of the buffer.
        if(iReadAheadPos>=iActual) getMoreBytes();

        return readAhead[iReadAheadPos++];
    }

    /**
     * reads an float (16 bit) from the file
     * @return the float value
     */
    public short readShort()
    {
        byte b1 = getByte();
        byte b2 = getByte();
        if(byteOrd==intelOrdering)
        {
            return (short)(intBits(b2,8)|intBits(b1,0));
        }
        else
        {
            return (short)(intBits(b1,8)|intBits(b2,0));
        }
    }

    /**
     * checks if more data is available
     * @return true if more data is available, otherwise false
     */
    public boolean moreAvailable()
    {
        if(bError==false)
        {
            if(iReadAheadPos<iActual) return true;
            try
            {
                if(input.available()!=0) return true;
            }
            catch(Exception e)
            {
                e.printStackTrace();
            }
        }
        return false; // theres an error
    }

    /**
     * reads more bytes from the file if the buffer is empty
     */
    protected void getMoreBytes()
    {
        try
        {
            iActual = Math.max(iReadAheadSize,input.available());
            input.read(readAhead,0,iActual);
            iReadAheadPos=0;
        }
        catch(Exception e)
        {
            e.printStackTrace();
            bError=true;
        }
    }
}
